"""
   Copyright 2021 Dugal Harris - dugalh@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import logging
from contextlib import contextmanager
import itertools
import json
import os
import pathlib
import sys
import time
from threading import Thread
from typing import Tuple, Optional

import ee
import rasterio as rio
import requests
from geedim.enums import ResamplingMethod
from rasterio.warp import transform_geom
from rasterio.env import GDALVersion
from rasterio.windows import Window
from requests.adapters import HTTPAdapter, Retry
import numpy as np
from tqdm.auto import tqdm

if '__file__' in globals():
    root_path = pathlib.Path(__file__).absolute().parents[1]
else:
    root_path = pathlib.Path(os.getcwd())

_GDAL_AT_LEAST_35 = GDALVersion.runtime().at_least("3.5")

def Initialize(opt_url: Optional[str]='https://earthengine-highvolume.googleapis.com', **kwargs):
    """
    Initialise Earth Engine.

    Credentials will be read from the `EE_SERVICE_ACC_PRIVATE_KEY` environment variable if it exists
    (useful for integrating with e.g. GitHub actions).

    .. note::

        Earth Engine recommends using the `high volume endpoint` for applications like ``geedim``.
        See `the docs <https://developers.google.com/earth-engine/cloud/highvolume>`_ for more information.

    Parameters
    ----------
    opt_url: str
        The Earth Engine endpoint to use.  If ``None``, the default is used.
    kwargs
        Optional arguments to pass to `ee.Initialize`.
    """

    if not ee.data._credentials:
        # Adpated from https://gis.stackexchange.com/questions/380664/how-to-de-authenticate-from-earth-engine-api.
        env_key = 'EE_SERVICE_ACC_PRIVATE_KEY'

        if env_key in os.environ:
            # authenticate with service account
            key_dict = json.loads(os.environ[env_key])
            credentials = ee.ServiceAccountCredentials(key_dict['client_email'], key_data=key_dict['private_key'])
            ee.Initialize(credentials, opt_url=opt_url, **kwargs)
        else:
            ee.Initialize(opt_url=opt_url, **kwargs)


def singleton(cls):
    """ Class decorator to make it a singleton. """
    instances = {}

    def getinstance() -> cls:
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]

    return getinstance


def split_id(image_id: str) -> Tuple[str, str]:
    """
    Split Earth Engine image ID into collection and index components.

    Parameters
    ----------
    image_id: str
        Earth engine image ID.

    Returns
    -------
    tuple(str, str)
        A tuple of strings: (collection name, image index).
    """
    if not image_id:
        return None, None
    index = image_id.split("/")[-1]
    ee_coll_name = "/".join(image_id.split("/")[:-1])
    return ee_coll_name, index


@contextmanager
def suppress_rio_warn_logs():
    """ A context manager that sets the `rasterio` logging level to ERROR, then returns it to its original value. """
    try:
        # GEE sets GeoTIFF `colorinterp` tags incorrectly. This suppresses `rasterio` warning relating to this:
        # 'Sum of Photometric type-related color channels and ExtraSamples doesn't match SamplesPerPixel'
        rio_level = logging.getLogger('rasterio').getEffectiveLevel()
        logging.getLogger('rasterio').setLevel(logging.ERROR)
        yield
    finally:
        logging.getLogger('rasterio').setLevel(rio_level)


def get_bounds(filename: pathlib.Path, expand: float = 5):
    """
    Get a geojson polygon representing the bounds of an image.

    Parameters
    ----------
    filename: str, pathlib.Path
        Path of the image file whose bounds to find.
    expand : int, optional
        Percentage (0-100) by which to expand the bounds (default: 5).

    Returns
    -------
    dict
        Geojson polygon.
    """
    with suppress_rio_warn_logs(), rio.Env(GTIFF_FORCE_RGBA=False), rio.open(filename) as im:
        bbox = im.bounds
        if (im.crs.linear_units == "metre") and (expand > 0):  # expand the bounding box
            expand_x = (bbox.right - bbox.left) * expand / 100.0
            expand_y = (bbox.top - bbox.bottom) * expand / 100.0
            bbox_expand = rio.coords.BoundingBox(
                bbox.left - expand_x, bbox.bottom - expand_y, bbox.right + expand_x, bbox.top + expand_y
            )
        else:
            bbox_expand = bbox

        coordinates = [
            [bbox_expand.right, bbox_expand.bottom], [bbox_expand.right, bbox_expand.top],
            [bbox_expand.left, bbox_expand.top], [bbox_expand.left, bbox_expand.bottom],
            [bbox_expand.right, bbox_expand.bottom],
        ]

        bbox_expand_dict = dict(type="Polygon", coordinates=[coordinates])
        src_bbox_wgs84 = transform_geom(im.crs, "WGS84", bbox_expand_dict)  # convert to WGS84 geojson
    return src_bbox_wgs84


def get_projection(image, min_scale=True):
    """
    Get the min/max scale projection of image bands.  Server side - no calls to getInfo().
    Adapted from from https://github.com/gee-community/gee_tools, MIT license.

    Parameters
    ----------
    image : ee.Image
            Image whose min/max projection to retrieve.
    min_scale: bool, optional
         Retrieve the projection corresponding to the band with the minimum (True) or maximum (False) scale.
         (default: True)

    Returns
    -------
    ee.Projection
        Requested projection.
    """
    if not isinstance(image, ee.Image):
        raise TypeError('image is not an instance of ee.Image')

    bands = image.bandNames()

    compare = ee.Number.lte if min_scale else ee.Number.gte
    init_proj = image.select(0).projection()

    def compare_scale(name, prev_proj):
        """Server side comparison of band scales"""
        prev_proj = ee.Projection(prev_proj)
        prev_scale = prev_proj.nominalScale()

        curr_proj = image.select([name]).projection()
        curr_scale = ee.Number(curr_proj.nominalScale())

        condition = compare(curr_scale, prev_scale)
        comp_proj = ee.Algorithms.If(condition, curr_proj, prev_proj)
        return ee.Projection(comp_proj)

    return ee.Projection(bands.iterate(compare_scale, init_proj))


class Spinner(Thread):

    def __init__(self, label='', interval=0.2, leave=True, **kwargs):
        """
        Thread sub-class to run a non-blocking spinner.

        Parameters
        ----------
        label: str, optional
            Prepend spinner with this label.
        interval: float, optional
            Spinner update interval (s).
        leave: optional, bool, str
            What to do with the spinner display on stop():
                False: clear the label + spinner.
                True:  leave the label + spinner as is.
                <string message>: print this message in place of the spinner
        kwargs: optional
            Additional kwargs to pass to Thread.__init__()
        """
        Thread.__init__(self, **kwargs)
        self._label = label
        self._interval = interval
        self._run = True
        self._leave = leave
        self._file = sys.stderr

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        self.join()

    def run(self):
        """ Run the spinner thread. """
        cursors_it = itertools.cycle('/-\|')

        while self._run:
            cursor = next(cursors_it)
            tqdm.write('\r' + self._label + cursor, file=self._file, end='')
            self._file.flush()
            time.sleep(self._interval)

        if self._leave == True:
            tqdm.write('', file=self._file, end='\n')
        elif self._leave == False:
            tqdm.write('\r', file=self._file, end='')
        elif isinstance(self._leave, str):
            tqdm.write('\r' + self._label + self._leave + ' ', file=self._file, end='\n')
        self._file.flush()

    def start(self):
        """ Start the spinner thread. """
        self._run = True
        Thread.start(self)

    def stop(self):
        """ Stop the spinner thread. """
        self._run = False


def resample(ee_image: ee.Image, method: ResamplingMethod) -> ee.Image:
    """
    Resample an ee.Image. Extends ee.Image.resample by only resampling when the image has a fixed projection, and by
    providing an additional 'average' method for downsampling.  This logic is performed server-side.

    See https://developers.google.com/earth-engine/guides/resample for more info.

    Parameters
    ----------
    ee_image: ee.Image
        Image to resample.
    method: ResamplingMethod
        Resampling method to use.

    Returns
    -------
    ee_image: ee.Image
        Resampled image.
    """
    # TODO : use STAC to only resample continuous qty type bands
    method = ResamplingMethod(method)
    if method == ResamplingMethod.near:
        return ee_image

    # resample the image, if it has a fixed projection
    proj = ee_image.select(0).projection()
    has_fixed_proj = proj.crs().compareTo('EPSG:4326').neq(0).Or(proj.nominalScale().toInt64().neq(111319))

    def _resample(ee_image: ee.Image) -> ee.Image:
        """ Resample the given image, allowing for additional 'average' method. """
        if method == ResamplingMethod.average:
            return ee_image.reduceResolution(reducer=ee.Reducer.mean(), maxPixels=1024)
        else:
            return ee_image.resample(method.value)

    return ee.Image(ee.Algorithms.If(has_fixed_proj, _resample(ee_image), ee_image))


def retry_session(
    retries: int = 3, backoff_factor: float = 0.3, status_forcelist: Tuple = (429, 500, 502, 503, 504),
    session: requests.Session = None
) -> requests.Session:
    """ requests session configured for retries. """
    session = session or requests.Session()
    retry = Retry(
        total=retries, read=retries, connect=retries, backoff_factor=backoff_factor, status_forcelist=status_forcelist
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def expand_window_to_grid(win: Window, expand_pixels: Tuple[int, int] = (0, 0)) -> Window:
    """
    Expand rasterio window extents to the nearest whole numbers i.e. for ``expand_pixels`` >= (0, 0), it will return a
    window that contains the original extents.

    Parameters
    ----------
    win: rasterio.windows.Window
        Window to expand.
    expand_pixels: tuple, optional
        Tuple specifying the number of (rows, columns) pixels to expand the window by.

    Returns
    -------
    rasterio.windows.Window
        Expanded window.
    """
    col_off, col_frac = np.divmod(win.col_off - expand_pixels[1], 1)
    row_off, row_frac = np.divmod(win.row_off - expand_pixels[0], 1)
    width = np.ceil(win.width + 2 * expand_pixels[1] + col_frac)
    height = np.ceil(win.height + 2 * expand_pixels[0] + row_frac)
    exp_win = Window(int(col_off), int(row_off), int(width), int(height))
    return exp_win


def rio_crs(crs: str) -> str:
    """ Convert a GEE CRS string to a rasterio compatible CRS string. """
    if crs == 'SR-ORG:6974':
        # This is a workaround for https://issuetracker.google.com/issues/194561313, that replaces the alleged GEE
        # SR-ORG:6974 with actual (rasterio) SR-ORG:6842. WKT taken from
        # https://spatialreference.org/ref/sr-org/modis-sinusoidal/html/.
        crs = """PROJCS["Sinusoidal",
GEOGCS["GCS_Undefined",
    DATUM["Undefined",
        SPHEROID["User_Defined_Spheroid",6371007.181,0.0]],
    PRIMEM["Greenwich",0.0],
    UNIT["Degree",0.0174532925199433]],
PROJECTION["Sinusoidal"],
PARAMETER["False_Easting",0.0],
PARAMETER["False_Northing",0.0],
PARAMETER["Central_Meridian",0.0],
UNIT["Meter",1.0]]"""
    return crs


def asset_id(image_id: str, folder: str = None):
    """
    Convert an EE image ID and EE asset project into an EE asset ID.

    If ``folder`` is not specified, ``image_id`` is returned as is. Otherwise, ``image_id`` is converted to a name by
    changing forward slashes to dashes, ``folder`` is split into <root folder> and <sub folder> sections, and a string
    is returned with EE asset ID format:

        projects/<root folder>/assets/<sub folder(s)>/<image_name>.
    """
    if not folder:
        return image_id
    im_name = image_id.replace('/', '-')
    folder = pathlib.PurePosixPath(folder)
    cloud_folder = pathlib.PurePosixPath(folder.parts[0])
    asset_path = pathlib.PurePosixPath('/'.join(folder.parts[1:])).joinpath(im_name)
    return f'projects/{str(cloud_folder)}/assets/{str(asset_path)}'



"""
   Copyright 2021 Dugal Harris - dugalh@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

##
import logging
import os
import pathlib
import threading
import time
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from itertools import product
from typing import Tuple, Dict, List, Union, Iterator, Optional

import ee
import numpy as np
import rasterio as rio
from rasterio.crs import CRS
from rasterio.enums import Resampling as RioResampling
from rasterio import windows, features, warp
from tqdm import TqdmWarning
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from geedim import utils
from geedim.enums import ResamplingMethod, ExportType
from geedim.stac import StacCatalog, StacItem
from geedim.tile import Tile

logger = logging.getLogger(__name__)

supported_dtypes = ['uint8', 'uint16', 'uint32', 'int8', 'int16', 'int32', 'float32', 'float64']
""" Supported image data types for downloading/exporting. """
# Note:
# - while gdal >= 3.5 supports *int64 data type, there is a rasterio bug retrieving the *int64 nodata value,
# so geedim will not support these types for now.
# - the ordering of the list above is relevant to the auto dtype and should be: unsigned ints smallest - largest,
# signed ints smallest to largest, float types smallest to largest.


class BaseImage:
    _float_nodata = float('nan')
    _desc_width = 50
    _default_resampling = ResamplingMethod.near
    _ee_max_tile_size = 32
    _ee_max_tile_dim = 10000
    _default_export_type = ExportType.drive

    def __init__(self, ee_image: ee.Image):
        """
        A class for describing and downloading an Earth Engine image.

        Allows download and export without size limits, and provides client-side access to image properties.

        Parameters
        ----------
        ee_image: ee.Image
            Earth Engine image to encapsulate.
        """
        if not isinstance(ee_image, ee.Image):
            raise TypeError('`ee_image` must be an instance of ee.Image.')
        self._ee_image = ee_image
        self.__ee_info = None
        self._id = None
        self.__min_projection = None
        self._min_dtype = None

    @classmethod
    def from_id(cls, image_id: str) -> 'BaseImage':
        """
        Create a BaseImage instance from an Earth Engine image ID.

        Parameters
        ----------
        image_id : str
           ID of earth engine image to wrap.

        Returns
        -------
        BaseImage
            BaseImage instance.
        """
        ee_image = ee.Image(image_id)
        gd_image = cls(ee_image)
        gd_image._id = image_id  # set the id attribute from image_id (avoids a call to getInfo() for .id property)
        return gd_image

    @property
    def _ee_info(self) -> Dict:
        """ Earth Engine image metadata. """
        if self.__ee_info is None:
            self.__ee_info = self._ee_image.getInfo()
        return self.__ee_info

    @property
    def _min_projection(self) -> Dict:
        """ Projection information corresponding to the minimum scale band. """
        if not self.__min_projection:
            self.__min_projection = self._get_projection(self._ee_info, min_scale=True)
        return self.__min_projection

    @property
    def _stac(self) -> Optional[StacItem]:
        """ Image STAC info.  None if there is no Earth Engine STAC entry for the image / image's collection. """
        return StacCatalog().get_item(self.id)

    @property
    def ee_image(self) -> ee.Image:
        """ Encapsulated Earth Engine image. """
        return self._ee_image

    @ee_image.setter
    def ee_image(self, value: ee.Image):
        self.__ee_info = None
        self.__min_projection = None
        self._min_dtype = None
        self._ee_image = value

    @property
    def id(self) -> Optional[str]:
        """ Earth Engine image ID.  None if the image ``system:id`` property is not set. """
        if self._id:  # avoid a call to getInfo() if _id is set
            return self._id
        else:
            return self._ee_info['id'] if 'id' in self._ee_info else None

    @property
    def date(self) -> Optional[datetime]:
        """ Image capture date & time.  None if the image ``system:time_start`` property is not set. """
        if 'system:time_start' in self.properties:
            return datetime.utcfromtimestamp(self.properties['system:time_start'] / 1000)
        else:
            return None

    @property
    def name(self) -> Optional[str]:
        """ Image name (the :attr:`id` with slashes replaced by dashes). """
        return self.id.replace('/', '-') if self.id else None

    @property
    def crs(self) -> Optional[str]:
        """
        Image CRS corresponding to minimum scale band, as an EPSG string. None if the image has no fixed projection.
        """
        return self._min_projection['crs']

    @property
    def scale(self) -> Optional[float]:
        """ Minimum scale (m) of image bands. None if the image has no fixed projection. """
        return self._min_projection['scale']

    @property
    def footprint(self) -> Optional[Dict]:
        """ Geojson polygon of the image extent.  None if the image is a composite. """
        if ('properties' not in self._ee_info) or ('system:footprint' not in self._ee_info['properties']):
            return None
        footprint = self._ee_info['properties']['system:footprint']

        if ('type' in footprint) and ('coordinates' in footprint) and footprint['type'] == 'LinearRing':
            # convert LinearRing to simple Polygon
            # this is necessary to make footprint compatible as a region in other geedim methods e.g.
            # MaskedCollection.search or BaseImage.download.
            footprint['coordinates'] = [footprint['coordinates']]
            footprint['type'] = 'Polygon'

        return self._ee_info['properties']['system:footprint']

    @property
    def shape(self) -> Optional[Tuple[int, int]]:
        """ (row, column) pixel dimensions of the minimum scale band. None if the image has no fixed projection. """
        return self._min_projection['shape']

    @property
    def count(self) -> int:
        """ Number of image bands. """
        return len(self._ee_info['bands']) if 'bands' in self._ee_info else None

    @property
    def transform(self) -> Optional[rio.Affine]:
        """ Geo-transform of the minimum scale band. None if the image has no fixed projection. """
        return self._min_projection['transform']

    @property
    def dtype(self) -> str:
        """ Minimum size data type able to represent the image values. """
        if not self._min_dtype:
            self._min_dtype = self._get_min_dtype(self._ee_info)
        return self._min_dtype

    @property
    def size(self) -> Optional[int]:
        """ Image size (bytes).  None if the image has no fixed projection. """
        if not self.shape:
            return None
        dtype_size = np.dtype(self.dtype).itemsize
        return self.shape[0] * self.shape[1] * self.count * dtype_size

    @property
    def has_fixed_projection(self) -> bool:
        """ True if the image has a fixed projection, otherwise False. """
        return self.scale is not None

    @property
    def properties(self) -> Dict:
        """ Earth Engine image properties. """
        return self._ee_info['properties'] if 'properties' in self._ee_info else {}

    @property
    def band_properties(self) -> List[Dict]:
        """ Merged STAC and Earth Engine band properties. """
        return self._get_band_properties()

    @property
    def refl_bands(self) -> Optional[List[str]]:
        """ List of spectral / reflectance bands, if any. """
        if not self._stac:
            return None
        return [bname for bname, bdict in self._stac.band_props.items() if 'center_wavelength' in bdict]

    @property
    def profile(self) -> Dict:
        """ Rasterio image profile. """
        return dict(
            crs=self.crs, transform=self.transform, width=self.shape[1], height=self.shape[0], dtype=self.dtype,
            count=self.count
        )

    @property
    def bounded(self) -> bool:
        """ True if the image is bounded, otherwise False. """
        # TODO: an unbounded region could also have these bounds
        unbounded_bounds = (-180, -90, 180, 90)
        return (self.footprint is not None) and (features.bounds(self.footprint) != unbounded_bounds)

    @staticmethod
    def _get_projection(ee_info: Dict, min_scale=True) -> Dict:
        """
        Return the projection information corresponding to the min/max scale band of a given Earth Engine image info
        dictionary.
        """
        projection_info = dict(crs=None, transform=None, shape=None, scale=None)
        if 'bands' in ee_info:
            # get scale & crs corresponding to min/max scale band
            scales = np.array([abs(bd['crs_transform'][0]) for bd in ee_info['bands']])
            crss = np.array([bd['crs'] for bd in ee_info['bands']])
            fixed_idx = (crss != 'EPSG:4326') | (scales != 1)
            if sum(fixed_idx) > 0:
                idx = np.argmin(scales[fixed_idx]) if min_scale else np.argmax(scales[fixed_idx])
                band_info = np.array(ee_info['bands'])[fixed_idx][idx]
                projection_info['scale'] = abs(band_info['crs_transform'][0])
                projection_info['crs'] = band_info['crs']
                if 'dimensions' in band_info:
                    projection_info['shape'] = band_info['dimensions'][::-1]
                projection_info['transform'] = rio.Affine(*band_info['crs_transform'])
                if ('origin' in band_info) and not np.any(np.isnan(band_info['origin'])):
                    projection_info['transform'] *= rio.Affine.translation(*band_info['origin'])
        return projection_info

    @staticmethod
    def _get_min_dtype(ee_info: Dict) -> str:
        """ Return the minimum size rasterio data type corresponding to a given Earth Engine image info dictionary. """
        def get_min_int_dtype(band_info: List[Dict]) -> Optional[str]:
            """ Return the minimum integer dtype for the given band information. """
            # create a list of integer min/max values
            int_minmax = [
                minmax for band_dict in band_info if band_dict['data_type']['precision'] == 'int'
                for minmax in (int(band_dict['data_type']['min']), int(band_dict['data_type']['max']))
            ]  # yapf: disable

            if len(int_minmax) == 0:
                return None
            min_value = np.nanmin(int_minmax)
            max_value = np.nanmax(int_minmax)

            for dtype in supported_dtypes[:-2]:
                if (min_value >= np.iinfo(dtype).min) and (max_value <= np.iinfo(dtype).max):
                    return dtype
            return 'float64'

        # TODO: from gdal = 3.5, there is support *int64 in geotiffs, rasterio 1.3.0 had some nodata issues with this,
        #  int64 support should be added when these issues are resolved
        dtype = None
        if 'bands' in ee_info:
            precisions = np.array([bd['data_type']['precision'] for bd in ee_info['bands']])
            if any(precisions == 'double'):
                dtype = 'float64'
            elif any(precisions == 'float'):
                # if there are >= 32 integer bits, use float64 to accommodate them, otherwise float32
                int_dtype = get_min_int_dtype(ee_info['bands'])
                dtype = 'float32' if (not int_dtype or np.dtype(int_dtype).itemsize < 4) else 'float64'
            else:
                dtype = get_min_int_dtype(ee_info['bands'])
        return dtype

    @staticmethod
    def _str_format_size(byte_size: float, units=['bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB']) -> str:
        """
        Returns a human readable string representation of a given byte size.
        Adapted from https://stackoverflow.com/questions/1094841/get-human-readable-version-of-file-size.
        """
        if byte_size < 1000:
            return f'{byte_size:.2f} {units[0]}'
        else:
            return BaseImage._str_format_size(byte_size / 1000, units[1:])

    @staticmethod
    def _convert_dtype(ee_image: ee.Image, dtype: str) -> ee.Image:
        """ Convert the data type of an Earth Engine image to a specified type. """
        conv_dict = dict(
            float32=ee.Image.toFloat, float64=ee.Image.toDouble, uint8=ee.Image.toUint8, int8=ee.Image.toInt8,
            uint16=ee.Image.toUint16, int16=ee.Image.toInt16, uint32=ee.Image.toUint32, int32=ee.Image.toInt32,
        )
        if dtype not in conv_dict:
            raise TypeError(f'Unsupported dtype: {dtype}')

        return conv_dict[dtype](ee_image)

    def _get_band_properties(self) -> List[Dict]:
        """ Merge Earth Engine and STAC band properties for this image. """
        band_ids = [bd['id'] for bd in self._ee_info['bands']]
        if self._stac:
            stac_bands_props = self._stac.band_props
            band_props = [stac_bands_props[bid] if bid in stac_bands_props else dict(name=bid) for bid in band_ids]
        else:  # just use the image band IDs
            band_props = [dict(name=bid) for bid in band_ids]
        return band_props

    @staticmethod
    def _scale_offset(ee_image: ee.Image, band_properties: List[Dict]) -> ee.Image:
        """
        Apply any STAC band scales and offsets to an EE image.

        Applying STAC scales and offsets will convert pixel values from representative integer ranges to the floating
        point ranges of physical quantities (e.g. surface reflectance in the range 0-1).

        Parameters
        ----------
        ee_image: ee.Image
            The EE image to scale and offset.
        band_properties: list(dict)
            A list of dictionaries specifying band names and corresponding scale and or offset values e.g.
            :attr:`BaseImage.band_properties`.

        Returns
        -------
        ee.Image
            The scaled and offset image, with the same bands as ``ee_image``, and in the same order.
        """
        if band_properties is None:
            logger.warning('Cannot scale and offset this image, there is no STAC band information.')
            return ee_image

        # make band scale and offset dicts
        scale_dict = {bp['name']: bp['scale'] if 'scale' in bp else 1. for bp in band_properties}
        offset_dict = {bp['name']: bp['offset'] if 'offset' in bp else 0. for bp in band_properties}

        if all([s == 1 for s in scale_dict.values()]) and all([o == 0 for o in offset_dict.values()]):
            # all scales==1 and all offsets==0
            return ee_image

        adj_bands = ee_image.bandNames().filter(ee.Filter.inList('item', list(scale_dict.keys())))
        non_adj_bands = ee_image.bandNames().removeAll(adj_bands)

        # apply the scales and offsets
        scale_im = ee.Dictionary(scale_dict).toImage().select(adj_bands)
        offset_im = ee.Dictionary(offset_dict).toImage().select(adj_bands)
        adj_im = ee_image.select(adj_bands).multiply(scale_im).add(offset_im)

        # Typically, `band_properties` will contain items for every band, but there may be cases where there
        # are additional image bands in `ee_image` not in `band_properties`. Here, these additional bands are added
        # back to the adjusted image.
        adj_im = adj_im.addBands(ee_image.select(non_adj_bands))
        adj_im = adj_im.select(ee_image.bandNames())  # keep bands in original order
        # copy original ee_image properties and return
        return ee.Image(adj_im.copyProperties(ee_image, ee_image.propertyNames()))

    def _prepare_for_export(
        self, crs: str = None, crs_transform: Tuple[float] = None, shape: Tuple[int, int] = None, region: Dict =
        None, scale: float = None, resampling: ResamplingMethod = _default_resampling, dtype: str = None,
        scale_offset: bool = False
    ) -> 'BaseImage':
        """
        Prepare the encapsulated image for export/download.  Will reproject, resample, clip and convert the image
        according to the provided parameters.

        Export bounds and resolution can be specified with ``region`` and ``scale`` / ``shape``, or ``crs_transform``
        and ``shape``.  If no bounds are specified (with either ``region``, or ``crs_transform`` & ``shape``), the
        entire image granule is exported.

        When ``crs``, ``scale``, ``crs_transform`` & ``shape`` are not specified, the pixel grid of the exported
        image will coincide with that of the encapsulated image.

        Parameters
        ----------
        crs : str, optional
            WKT or EPSG specification of CRS to export to.  Where image bands have different CRSs, all are
            re-projected to this CRS. Defaults to use the CRS of the minimum scale band if available.
        crs_transform: tuple of float, list of float, rio.Affine, optional
            List of 6 numbers specifying an affine transform in the specified CRS.  In row-major order:
            [xScale, xShearing, xTranslation, yShearing, yScale, yTranslation].  All bands are re-projected to
            this transform.
        shape: tuple of int, optional
            (height, width) dimensions to export (pixels).
        region : dict, geojson, ee.Geometry, optional
            Region of interest (WGS84) to export.  Defaults to the image footprint, when available.
        scale : float, optional
            Pixel scale (m) to export to.  Where image bands have different scales, all are re-projected to this scale.
            Ignored if ``crs`` and ``crs_transform`` are specified.  Defaults to use the minimum scale of
            image bands if available.
        resampling : ResamplingMethod, optional
            Resampling method - see :class:`~geedim.enums.ResamplingMethod` for available options.
        dtype: str, optional
           Convert to this data type (`uint8`, `int8`, `uint16`, `int16`, `uint32`, `int32`, `float32` or
           `float64`). Defaults to auto select a minimal type that can represent the range of pixel values.
        scale_offset: bool, optional
            Whether to apply any EE band scales and offsets to the image.

        Returns
        -------
        BaseImage
            Prepared image.
        """

        # Check for parameter combination errors.
        # This necessarily duplicates some of what is done in ee.Image.getDownloadURL, so that errors are raised
        # before tile creation & download.
        if (not crs or not region or not scale) and (not crs or not crs_transform or not shape):
            if not self.has_fixed_projection:
                # if the image has no fixed projection, either crs, region, & scale; or crs, crs_transform and shape
                # must be specified
                raise ValueError(
                    f'This image does not have a fixed projection, you need to specify a crs, region & scale; or a '
                    f'crs, crs_transform & shape.'
                )

        if (not self.bounded) and (not region and (not crs or not crs_transform or not shape)):
            # if the image has no footprint (i.e. it is 'unbounded'), either region; or crs, crs_transform and shape
            # must be specified
            raise ValueError(
                f'This image is unbounded, you need to specify a region; or a crs, crs_transform and '
                f'shape.'
            )

        if self.crs == 'EPSG:4326' and not scale and not shape:
            # If the image is in EPSG:4326, either scale (in meters); or shape must be specified.
            # Note that ee.Image.prepare_for_export() expects a scale in meters, but if the image is EPSG:4326,
            # the default scale is in degrees.
            raise ValueError(f'This image is in EPSG:4326, you need to specify a scale (in meters); or a shape.')

        if scale and shape:
            # This error is raised in later calls to ee.Image.getInfo(), but is neater to raise here first
            raise ValueError('You can specify one of scale or shape, but not both.')

        # perform image scale/offset, dtype and resampling operations
        ee_image = self.ee_image
        if scale_offset:
            ee_image = self._scale_offset(ee_image, self.band_properties)
            im_dtype = 'float64'
        else:
            im_dtype = self.dtype

        resampling = ResamplingMethod(resampling)
        if resampling != self._default_resampling:
            if not self.has_fixed_projection:
                raise ValueError(
                    'This image has no fixed projection and cannot be resampled.  If this image is a composite, '
                    'you can resample the images used to create the composite.'
                )
            ee_image = utils.resample(ee_image, resampling)

        ee_image = self._convert_dtype(ee_image, dtype=dtype or im_dtype)

        # configure the export parameter values
        crs = crs or self.crs
        if not crs_transform and not shape:
            # if none of crs_transform, shape, region or scale are specified, then set region and scale to defaults
            region = region or self.footprint
            scale = scale or self.scale

            if (crs == self.crs) and (scale == self.scale):
                # if crs_transform & shape are not already specified, and crs & scale match this image's crs & scale,
                # then set crs_transform and shape to export on this image's (self) pixel grid
                if region == self.footprint:
                    crs_transform = self.transform
                    shape = self.shape
                else:
                    # find a crs_transform and shape that encompasses region
                    if isinstance(region, ee.Geometry):
                        region = region.getInfo()
                    region_crs = region['crs']['properties']['name'] if 'crs' in region else 'EPSG:4326'
                    region_bounds = warp.transform_bounds(region_crs, utils.rio_crs(crs), *features.bounds(region))
                    region_win = windows.from_bounds(*region_bounds, transform=self.transform)
                    region_win = utils.expand_window_to_grid(region_win)
                    crs_transform = self.transform * rio.Affine.translation(region_win.col_off, region_win.row_off)
                    shape = (region_win.height, region_win.width)

                # prevent exporting with region & scale now that crs_transform and shape are set
                region = None
                scale = None

        # create the export parameter dict
        crs_transform = crs_transform[:6] if crs_transform else None
        dimensions = shape[::-1] if shape else None
        export_kwargs = dict(
            crs=crs, crs_transform=crs_transform, dimensions=dimensions, region=region, scale=scale,
            fileFormat='GeoTIFF', filePerBand=False
        )
        # drop items with values==None
        export_kwargs = {k: v for k, v in export_kwargs.items() if v is not None}
        logger.debug(f'ee.Image.prepare_for_export() params: {export_kwargs}')
        ee_image, _ = ee_image.prepare_for_export(export_kwargs)
        return BaseImage(ee_image)

    def _prepare_for_download(self, set_nodata: bool = True, **kwargs) -> Tuple['BaseImage', Dict]:
        """
        Prepare the encapsulated image for tiled GeoTIFF download. Will reproject, resample, clip and convert the image
        according to the provided parameters.

        Returns the prepared image and a rasterio profile for the downloaded GeoTIFF.
        """
        # resample, convert, clip and reproject image according to download params
        exp_image = self._prepare_for_export(**kwargs)
        # see float nodata workaround note in Tile.download(...)
        nodata_dict = dict(
            float32=self._float_nodata,
            float64=self._float_nodata,
            uint8=0,
            int8=np.iinfo('int8').min,
            uint16=0,
            int16=np.iinfo('int16').min,
            uint32=0,
            int32=np.iinfo('int32').min,
        )  # yapf: disable
        nodata = nodata_dict[exp_image.dtype] if set_nodata else None
        profile = dict(
            driver='GTiff', dtype=exp_image.dtype, nodata=nodata, width=exp_image.shape[1], height=exp_image.shape[0],
            count=exp_image.count, crs=CRS.from_string(utils.rio_crs(exp_image.crs)), transform=exp_image.transform,
            compress='deflate', interleave='band', tiled=True, photometric='MINISBLACK',
        )
        # add BIGTIFF support if the uncompressed image is bigger than 4GB
        if exp_image.size >= 4e9:
            profile.update(bigtiff=True)
        return exp_image, profile

    @staticmethod
    def _get_tile_shape(
        exp_image: 'BaseImage', max_tile_size: Optional[float] = None, max_tile_dim: Optional[int] = None
    ) -> Tuple[Tuple[int, int], int]:  # yapf: disable
        """
        Return a tile shape and number of tiles for a given BaseImage, such that the tile shape satisfies GEE
        download limits, and is 'square-ish'.
        """
        # convert max_tile_size from MB to bytes & set to EE default if None
        if max_tile_size and (max_tile_size > BaseImage._ee_max_tile_size):
            raise ValueError(
                f'`max_tile_size` must be less than or equal to the Earth Engine download size limit of '
                f'{BaseImage._ee_max_tile_size} MB.'
            )
        max_tile_size = int((max_tile_size * (1 << 20)) if max_tile_size else (BaseImage._ee_max_tile_size << 20))
        if max_tile_dim and (max_tile_dim > BaseImage._ee_max_tile_dim):
            raise ValueError(
                f'`max_tile_dim` must be less than or equal to the Earth Engine download limit of '
                f'{BaseImage._ee_max_tile_size} pixels.'
            )
        max_tile_dim = max_tile_dim or BaseImage._ee_max_tile_dim   # set max_tile_dim to EE default if None

        # find the total number of tiles the image must be divided into to satisfy max_tile_size
        image_shape = np.array(exp_image.shape, dtype='int64')
        dtype_size = np.dtype(exp_image.dtype).itemsize
        image_size = exp_image.size
        if exp_image.dtype.endswith('int8'):
            # workaround for GEE overestimate of *int8 dtype download sizes
            dtype_size *= 2
            image_size *= 2

        pixel_size = dtype_size * exp_image.count

        num_tile_shape = np.array([1, 1], dtype='int64')
        tile_size = image_size
        tile_shape = image_shape
        while tile_size >= max_tile_size:
            div_axis = np.argmax(tile_shape)
            num_tile_shape[div_axis] += 1  # increase the num tiles down the longest dimension of tile_shape
            tile_shape = np.ceil(image_shape / num_tile_shape).astype('int64')
            tile_size = tile_shape[0] * tile_shape[1] * pixel_size

        tile_shape[tile_shape > max_tile_dim] = max_tile_dim
        num_tiles = int(np.product(np.ceil(image_shape / tile_shape)))
        tile_shape = tuple(tile_shape.tolist())
        return tile_shape, num_tiles

    @staticmethod
    def _build_overviews(dataset: rio.io.DatasetWriter, max_num_levels: int = 8, min_ovw_pixels: int = 256):
        """ Build internal overviews, downsampled by successive powers of 2, for an open rasterio dataset. """
        if dataset.closed:
            raise IOError('Image dataset is closed')

        # limit overviews so that the highest level has at least 2**8=256 pixels along the shortest dimension,
        # and so there are no more than 8 levels.
        max_ovw_levels = int(np.min(np.log2(dataset.shape)))
        min_level_shape_pow2 = int(np.log2(min_ovw_pixels))
        num_ovw_levels = np.min([max_num_levels, max_ovw_levels - min_level_shape_pow2])
        ovw_levels = [2**m for m in range(1, num_ovw_levels + 1)]
        dataset.build_overviews(ovw_levels, RioResampling.average)

    def _write_metadata(self, dataset: rio.io.DatasetWriter):
        """ Write Earth Engine and STAC metadata to an open rasterio dataset. """
        if dataset.closed:
            raise IOError('Image dataset is closed')

        # replace 'system:*' property keys with 'system-*', and remove footprint if its there
        properties = {k.replace(':', '-'):v for k, v in self.properties.items()}
        if 'system-footprint' in properties:
            properties.pop('system-footprint')
        dataset.update_tags(**properties)

        if self._stac and self._stac.license:
            dataset.update_tags(LICENSE=self._stac.license)

        def clean_text(text, width=80) -> str:
            """ Return a shortened tidied string. """
            if not isinstance(text, str):
                return text
            text = text.split('.')[0] if len(text) > width else text
            text = text.strip()
            text = '-\n' + text if len(text) > width else text
            return text

        # populate band metadata
        for band_i, band_dict in enumerate(self.band_properties):
            clean_band_dict = {k.replace(':', '-'): clean_text(v) for k, v in band_dict.items()}
            if 'name' in band_dict:
                dataset.set_band_description(band_i + 1, clean_band_dict['name'])
            dataset.update_tags(band_i + 1, **clean_band_dict)

    @staticmethod
    def _tiles(exp_image: 'BaseImage', tile_shape: Tuple[int, int]) -> Iterator[Tile]:
        """
        Iterator over downloadable image tiles.

        Divides an image into adjoining tiles no bigger than `tile_shape`.

        Parameters
        ----------
        exp_image: BaseImage
            Image to tile.
        tile_shape: Tuple[int, int]
            (row, column) tile shape to use (pixels). Use :meth:`BaseImage._get_tile_shape` to find a tile shape that
            satisfies the Earth Engine download limit for :param:`exp_image`.

        Yields
        -------
        Tile
            An image tile that can be downloaded.
        """

        # split the image up into tiles of at most `tile_shape` dimension
        image_shape = exp_image.shape
        start_range = product(range(0, image_shape[0], tile_shape[0]), range(0, image_shape[1], tile_shape[1]))
        for tile_start in start_range:
            tile_stop = np.clip(np.add(tile_start, tile_shape), a_min=None, a_max=image_shape)
            clip_tile_shape = (tile_stop - tile_start).tolist()  # tolist is just to convert to native int
            tile_window = windows.Window(tile_start[1], tile_start[0], clip_tile_shape[1], clip_tile_shape[0])
            yield Tile(exp_image, tile_window)

    @staticmethod
    def monitor_export(task: ee.batch.Task, label: str = None):
        """
        Monitor and display the progress of an export task.

        Parameters
        ----------
        task : ee.batch.Task
            Earth Engine task to monitor (as returned by :meth:`export`).
        label: str, optional
            Optional label for progress display.  Defaults to task description.
        """
        pause = 0.1
        status = ee.data.getOperation(task.name)

        if label is None:
            label = status["metadata"]["description"]
        label = label if (len(label) < BaseImage._desc_width) else f'*{label[-BaseImage._desc_width:]}'

        # poll EE until the export preparation is complete
        with utils.Spinner(label=f'Preparing {label}: ', leave='done'):
            while ('done' not in status) or (not status['done']):
                time.sleep(5 * pause)
                status = ee.data.getOperation(task.name)
                if 'progress' in status['metadata']:
                    break
                elif status['metadata']['state'] == 'FAILED':
                    raise IOError(f"Export failed \n{status}")

        # wait for export to complete, displaying a progress bar
        bar_format = '{desc}: |{bar}| [{percentage:5.1f}%] in {elapsed:>5s} (eta: {remaining:>5s})'
        with tqdm(desc=f'Exporting {label}', total=1, bar_format=bar_format, dynamic_ncols=True) as bar:
            while ('done' not in status) or (not status['done']):
                time.sleep(10 * pause)
                status = ee.data.getOperation(task.name)  # get task status
                bar.update(status['metadata']['progress'] - bar.n)

            if status['metadata']['state'] == 'SUCCEEDED':
                bar.update(1 - bar.n)
            else:
                raise IOError(f"Export failed \n{status}")

    def export(
        self, filename: str, type: ExportType = _default_export_type, folder: str = None, wait: bool = True, **kwargs
    ) -> ee.batch.Task:
        """
        Export the encapsulated image to Google Drive, Earth Engine asset or Google Cloud Storage.

        Export bounds and resolution can be specified with ``region`` and ``scale`` / ``shape``, or ``crs_transform``
        and ``shape``.  If no bounds are specified (with either ``region``, or ``crs_transform`` & ``shape``), the
        entire image granule is exported.

        When ``crs``, ``scale``, ``crs_transform`` & ``shape`` are not specified, the pixel grid of the exported
        image will coincide with that of the encapsulated image.

        Parameters
        ----------
        filename : str
            Name of the export task, and destination file or asset name.
        type : ExportType, optional
            Export type.
        folder : str, optional
            Google Drive folder (when ``type`` is :attr:`~geedim.enums.ExportType.drive`),
            Earth Engine asset project (when ``type`` is :attr:`~geedim.enums.ExportType.asset`),
            or Google Cloud Storage bucket (when ``type`` is :attr:`~geedim.enums.ExportType.cloud`), to export to.
            If ``type`` is :attr:`~geedim.enums.ExportType.asset` and ``folder`` is not specified, ``filename``
            should be a valid Earth Engine asset ID.
        wait : bool
            Wait for the export to complete before returning.
        crs : str, optional
            WKT or EPSG specification of CRS to export to.  Where image bands have different CRSs, all are
            re-projected to this CRS. Defaults to use the CRS of the minimum scale band if available.
        crs_transform: tuple of float, list of float, rio.Affine, optional
            List of 6 numbers specifying an affine transform in the specified CRS.  In row-major order:
            [xScale, xShearing, xTranslation, yShearing, yScale, yTranslation].  All bands are re-projected to
            this transform.
        shape: tuple of int, optional
            (height, width) dimensions to export (pixels).
        region : dict, geojson, ee.Geometry, optional
            Region of interest (WGS84) to export.  Defaults to the image footprint, when available.
        scale : float, optional
            Pixel scale (m) to export to.  Where image bands have different scales, all are re-projected to this scale.
            Ignored if ``crs`` and ``crs_transform`` are specified.  Defaults to use the minimum scale of
            image bands if available.
        resampling : ResamplingMethod, optional
            Resampling method - see :class:`~geedim.enums.ResamplingMethod` for available options.
        dtype: str, optional
           Convert to this data type (`uint8`, `int8`, `uint16`, `int16`, `uint32`, `int32`, `float32` or
           `float64`). Defaults to auto select a minimal type that can represent the range of pixel values.
        scale_offset: bool, optional
            Whether to apply any EE band scales and offsets to the image.

        Returns
        -------
        ee.batch.Task
            The Earth Engine export task, started if ``wait`` is False, or completed if ``wait`` is True.
        """

        exp_image = self._prepare_for_export(**kwargs)
        if exp_image.crs == 'SR-ORG:6974':
            logger.warning(
                'There is an earth engine bug exporting in SR-ORG:6974, you will need to edit the exported file to '
                'replace the CRS with SR-ORG:6842. See: https://issuetracker.google.com/issues/194561313.'
            )

        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug(f'Uncompressed size: {self._str_format_size(exp_image.size)}')

        # create export task and start
        type = ExportType(type)
        if type == ExportType.drive:
            task = ee.batch.Export.image.toDrive(
                image=exp_image.ee_image, description=filename[:100], folder=folder, fileNamePrefix=filename,
                maxPixels=1e9, formatOptions=dict(cloudOptimized=True),
            )
        elif type == ExportType.asset:
            # if folder is specified create an EE asset ID from it and filename,
            # else assume filename is a valid EE asset ID
            asset_id = utils.asset_id(filename, folder) if folder else filename
            # fix description for when filename is asset id with forward slashes
            description = filename.replace('/', '-')[:100]
            task = ee.batch.Export.image.toAsset(
                image=exp_image.ee_image, description=description, assetId=asset_id, maxPixels=1e9,
            )
        else:
            task = ee.batch.Export.image.toCloudStorage(
                image=exp_image.ee_image, description=filename[:100], bucket=folder, maxPixels=1e9,
                formatOptions=dict(cloudOptimized=True),
            )

        task.start()
        if wait:  # wait for completion
            self.monitor_export(task)
        return task

    def download(
        self, filename: Union[pathlib.Path, str], overwrite: bool = False, num_threads: Optional[int] = None,
        max_tile_size: Optional[float] = None, max_tile_dim: Optional[int] = None, **kwargs
    ):
        """
        Download the encapsulated image to a GeoTiff file.

        Images larger than the `Earth Engine size limit
        <https://developers.google.com/earth-engine/apidocs/ee-image-getdownloadurl>`_ are split and downloaded as
        separate tiles, then re-assembled into a single GeoTIFF.  Downloaded image files are populated with metadata
        from the Earth Engine image and STAC.

        Image bounds and resolution can be specified with ``region`` and ``scale`` / ``shape``, or ``crs_transform``
        and ``shape``.  If no bounds are specified (with either ``region``, or ``crs_transform`` & ``shape``), the
        entire image granule is downloaded.

        When ``crs``, ``scale``, ``crs_transform`` & ``shape`` are not specified, the pixel grid of the exported
        image will coincide with that of the encapsulated image.

        Parameters
        ----------
        filename: pathlib.Path, str
            Name of the destination file.
        overwrite : bool, optional
            Overwrite the destination file if it exists.
        num_threads: int, optional
            Number of tiles to download concurrently.  Defaults to a sensible auto value.
        max_tile_size: int, optional
            Maximum tile size (MB).  If None, defaults to the Earth Engine download size limit (32 MB).
        max_tile_dim: int, optional
            Maximum tile width/height (pixels).  If None, defaults to Earth Engine download limit (10000).
        crs : str, optional
            WKT or EPSG specification of CRS to export to.  Where image bands have different CRSs, all are
            re-projected to this CRS. Defaults to use the CRS of the minimum scale band if available.
        crs_transform: tuple of float, list of float, rio.Affine, optional
            List of 6 numbers specifying an affine transform in the specified CRS.  In row-major order:
            [xScale, xShearing, xTranslation, yShearing, yScale, yTranslation].  All bands are re-projected to
            this transform.
        shape: tuple of int, optional
            (height, width) dimensions to export (pixels).
        region : dict, geojson, ee.Geometry, optional
            Region of interest (WGS84) to export.  Defaults to the image footprint, when available.
        scale : float, optional
            Pixel scale (m) to export to.  Where image bands have different scales, all are re-projected to this scale.
            Ignored if ``crs`` and ``crs_transform`` are specified.  Defaults to use the minimum scale of
            image bands if available.
        resampling : ResamplingMethod, optional
            Resampling method - see :class:`~geedim.enums.ResamplingMethod` for available options.
        dtype: str, optional
           Convert to this data type (`uint8`, `int8`, `uint16`, `int16`, `uint32`, `int32`, `float32` or
           `float64`). Defaults to auto select a minimal type that can represent the range of pixel values.
        scale_offset: bool, optional
            Whether to apply any EE band scales and offsets to the image.
        """

        max_threads = num_threads or min(32, (os.cpu_count() or 1) + 4)
        out_lock = threading.Lock()
        filename = pathlib.Path(filename)
        if filename.exists():
            if overwrite:
                os.remove(filename)
            else:
                raise FileExistsError(f'{filename} exists')

        # prepare (resample, convert, reproject) the image for download
        exp_image, profile = self._prepare_for_download(**kwargs)

        # get the dimensions of an image tile that will satisfy GEE download limits
        tile_shape, num_tiles = self._get_tile_shape(exp_image, max_tile_size=max_tile_size, max_tile_dim=max_tile_dim)

        # find raw size of the download data (less than the actual download size as the image data is zipped in a
        # compressed geotiff)
        raw_download_size = exp_image.size
        if logger.getEffectiveLevel() <= logging.DEBUG:
            dtype_size = np.dtype(exp_image.dtype).itemsize
            raw_tile_size = tile_shape[0] * tile_shape[1] * exp_image.count * dtype_size
            logger.debug(f'{filename.name}:')
            logger.debug(f'Uncompressed size: {self._str_format_size(raw_download_size)}')
            logger.debug(f'Num. tiles: {num_tiles}')
            logger.debug(f'Tile shape: {tile_shape}')
            logger.debug(f'Tile size: {self._str_format_size(int(raw_tile_size))}')

        if raw_download_size > 1e9:
            # warn if the download is large (>1GB)
            logger.warning(
                f'Consider adjusting `region`, `scale` and/or `dtype` to reduce the {filename.name}'
                f' download size (raw: {self._str_format_size(raw_download_size)}).'
            )

        # configure the progress bar to monitor raw/uncompressed download size
        desc = filename.name if (len(filename.name) < self._desc_width) else f'...{filename.name[-self._desc_width:]}'
        bar_format = (
            '{desc}: |{bar}| {n_fmt}/{total_fmt} (raw) [{percentage:5.1f}%] in {elapsed:>5s} (eta: {remaining:>5s})'
        )
        bar = tqdm(
            desc=desc, total=raw_download_size, bar_format=bar_format, dynamic_ncols=True, unit_scale=True, unit='B'
        )

        session = utils.retry_session(5)
        warnings.filterwarnings('ignore', category=TqdmWarning)
        # redirect logging through tqdm
        redir_tqdm = logging_redirect_tqdm([logging.getLogger(__package__)], tqdm_class=type(bar))
        env = rio.Env(GDAL_NUM_THREADS='ALL_CPUs', GTIFF_FORCE_RGBA=False)
        with redir_tqdm, env, rio.open(filename, 'w', **profile) as out_ds, bar:

            def download_tile(tile):
                """Download a tile and write into the destination GeoTIFF. """
                tile_array = tile.download(session=session, bar=bar)
                with out_lock:
                    out_ds.write(tile_array, window=tile.window)

            with ThreadPoolExecutor(max_workers=max_threads) as executor:
                # Run the tile downloads in a thread pool
                tiles = self._tiles(exp_image, tile_shape=tile_shape)
                futures = [executor.submit(download_tile, tile) for tile in tiles]
                try:
                    for future in as_completed(futures):
                        future.result()
                except Exception as ex:
                    logger.info(f'Exception: {str(ex)}\nCancelling...')
                    executor.shutdown(wait=False)
                    raise ex

            bar.update(bar.total - bar.n)   # ensure the bar reaches 100%
            # populate GeoTIFF metadata and build overviews
            self._write_metadata(out_ds)
            self._build_overviews(out_ds)



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from geedim.collection import MaskedCollection
from geedim.enums import CloudMaskMethod, CompositeMethod, ResamplingMethod, ExportType
from geedim.mask import MaskedImage
from geedim.utils import Initialize



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# schema definitions for MaskedImage.from_id(), geedim <-> EE collection names, and search properties
from tabulate import tabulate
from textwrap import wrap
import geedim.mask

# yapf: disable
default_prop_schema = {
    'system:id': {'abbrev': 'ID', 'description': 'Earth Engine image id'},
    'system:time_start': {'abbrev': 'DATE', 'description': 'Image capture date/time (UTC)'},
    'FILL_PORTION': {'abbrev': 'FILL', 'description': 'Portion of valid pixels (%)'},
}

landsat_prop_schema = {
    'system:id': {'abbrev': 'ID', 'description': 'Earth Engine image id'},
    'system:time_start': {'abbrev': 'DATE', 'description': 'Image capture date/time (UTC)'},
    'FILL_PORTION': {'abbrev': 'FILL', 'description': 'Portion of valid pixels (%)'},
    'CLOUDLESS_PORTION': {'abbrev': 'CLOUDLESS', 'description': 'Portion of cloud/shadow free pixels (%)'},
    'GEOMETRIC_RMSE_MODEL': {'abbrev': 'GRMSE', 'description': 'Orthorectification RMSE (m)'},
    'SUN_AZIMUTH': {'abbrev': 'SAA', 'description': 'Solar azimuth angle (deg)'},
    'SUN_ELEVATION': {'abbrev': 'SEA', 'description': 'Solar elevation angle (deg)'}
}

s2_prop_schema = {
    'system:id': {'abbrev': 'ID', 'description': 'Earth Engine image id'},
    'system:time_start': {'abbrev': 'DATE', 'description': 'Image capture date/time (UTC)'},
    'FILL_PORTION': {'abbrev': 'FILL', 'description': 'Portion of valid pixels (%)'},
    'CLOUDLESS_PORTION': {'abbrev': 'CLOUDLESS', 'description': 'Portion of cloud/shadow free pixels (%)'},
    'RADIOMETRIC_QUALITY': {'abbrev': 'RADQ', 'description': 'Radiometric quality check'},
    'GEOMETRIC_QUALITY': {'abbrev': 'GEOMQ', 'description': 'Geometric quality check'},
    'MEAN_SOLAR_AZIMUTH_ANGLE': {'abbrev': 'SAA', 'description': 'Solar azimuth angle (deg)'},
    'MEAN_SOLAR_ZENITH_ANGLE': {'abbrev': 'SZA', 'description': 'Solar zenith angle (deg)'},
    'MEAN_INCIDENCE_AZIMUTH_ANGLE_B1': {'abbrev': 'VAA', 'description': 'View (B1) azimuth angle (deg)'},
    'MEAN_INCIDENCE_ZENITH_ANGLE_B1': {'abbrev': 'VZA', 'description': 'View (B1) zenith angle (deg)'}
}

collection_schema = {
    'LANDSAT/LT04/C02/T1_L2': {
        'gd_coll_name': 'l4-c2-l2',
        'prop_schema': landsat_prop_schema,
        'image_type': geedim.mask.LandsatImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LT04_C02_T1_L2',
        'description': 'Landsat 4, collection 2, tier 1, level 2 surface reflectance.'
    },
    'LANDSAT/LT05/C02/T1_L2': {
        'gd_coll_name': 'l5-c2-l2',
        'prop_schema': landsat_prop_schema,
        'image_type': geedim.mask.LandsatImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LT05_C02_T1_L2',
        'description': 'Landsat 5, collection 2, tier 1, level 2 surface reflectance.'
    },
    'LANDSAT/LE07/C02/T1_L2': {
        'gd_coll_name': 'l7-c2-l2',
        'prop_schema': landsat_prop_schema,
        'image_type': geedim.mask.LandsatImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LE07_C02_T1_L2',
        'description': 'Landsat 7, collection 2, tier 1, level 2 surface reflectance.'
    },
    'LANDSAT/LC08/C02/T1_L2': {
        'gd_coll_name': 'l8-c2-l2',
        'prop_schema': landsat_prop_schema,
        'image_type': geedim.mask.LandsatImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LC08_C02_T1_L2',
        'description': 'Landsat 8, collection 2, tier 1, level 2 surface reflectance.'
    },
    'LANDSAT/LC09/C02/T1_L2': {
        'gd_coll_name': 'l9-c2-l2',
        'prop_schema': landsat_prop_schema,
        'image_type': geedim.mask.LandsatImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/LANDSAT_LC09_C02_T1_L2',
        'description': 'Landsat 9, collection 2, tier 1, level 2 surface reflectance.'
    },
    'COPERNICUS/S2': {
        'gd_coll_name': 's2-toa',
        'prop_schema': s2_prop_schema,
        'image_type': geedim.mask.Sentinel2ToaClImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2',
        'description': 'Sentinel-2, level 1C, top of atmosphere reflectance.'
    },
    'COPERNICUS/S2_SR': {
        'gd_coll_name': 's2-sr',
        'prop_schema': s2_prop_schema,
        'image_type': geedim.mask.Sentinel2SrClImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2_SR',
        'description': 'Sentinel-2, level 2A, surface reflectance.'
    },
    'COPERNICUS/S2_HARMONIZED': {
        'gd_coll_name': 's2-toa-hm',
        'prop_schema': s2_prop_schema,
        'image_type': geedim.mask.Sentinel2ToaClImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2_HARMONIZED',
        'description': 'Harmonised Sentinel-2, level 1C, top of atmosphere reflectance.'
    },
    'COPERNICUS/S2_SR_HARMONIZED': {
        'gd_coll_name': 's2-sr-hm',
        'prop_schema': s2_prop_schema,
        'image_type': geedim.mask.Sentinel2SrClImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/COPERNICUS_S2_SR_HARMONIZED',
        'description': 'Harmonised Sentinel-2, level 2A, surface reflectance.'
    },
    'MODIS/006/MCD43A4': {
        'gd_coll_name': 'modis-nbar',
        'prop_schema': default_prop_schema,
        'image_type': geedim.mask.MaskedImage,
        'ee_url': 'https://developers.google.com/earth-engine/datasets/catalog/MODIS_006_MCD43A4',
        'description': 'MODIS nadir BRDF adjusted daily reflectance.'
    }
}
# yapf: enable

# Dict to convert from geedim to Earth Engine collection names
ee_to_gd = dict([(k, v['gd_coll_name']) for k, v in collection_schema.items()])

# Dict to convert from Earth Engine to geedim collection names
gd_to_ee = dict([(v['gd_coll_name'], k) for k, v in collection_schema.items()])

# "Two way" dict to convert Earth Engine to/from geedim collection names
coll_names = dict(**gd_to_ee, **ee_to_gd)

# A list of cloud/shadow mask supported EE collection names
cloud_coll_names = [k for k, v in collection_schema.items() if v['image_type'] != geedim.mask.MaskedImage]


def cli_cloud_coll_table() -> str:
    """ Return a table of cloud/shadow mask supported collections for use in CLI help strings. """
    headers = dict(gd_coll_name='geedim name', ee_coll_name='EE name')
    data = []
    for key, val in collection_schema.items():
        if val['image_type'] != geedim.mask.MaskedImage:
            data.append(dict(gd_coll_name=val['gd_coll_name'], ee_coll_name=key))
    return tabulate(data, headers=headers, tablefmt='rst')


def cloud_coll_table(descr_join='\n') -> str:
    """
    Return a table of cloud/shadow mask supported collections.
    * Use descr_join='\n' for github README friendly formatting.
    * Use descr_join='\n\n' for RTD/Sphinx friendly formatting.

    Instructions for adding cloud/shadow supported collections to CLI help and github README:
    * print(cli_cloud_coll_table()) and paste into cli.search() and  cli.config() command docstrings.
    * print(cloud_coll_table()) and paste into the README.
    * The equivalent RTD table is auto-generated in docs/conf.py.
    """
    headers = dict(ee_coll_name='EE name', descr='Description')
    data = []
    for key, val in collection_schema.items():
        if val['image_type'] != geedim.mask.MaskedImage:
            ee_coll_name = '\n'.join(wrap(f'`{key} <{val["ee_url"]}>`_', width=40))
            descr = descr_join.join(wrap(val['description'], width=60))   # for RTD multiline table
            data.append(dict(ee_coll_name=ee_coll_name, descr=descr))

    return tabulate(data, headers=headers, tablefmt='grid')



"""
    MIT License

    Copyright (c) 2017 Rodrigo E. Principe

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
import ee
"""
    This module contains Medoid related functionality copied from 'Google Earth Engine tools' under MIT license.
    See https://github.com/gee-community/gee_tools.
"""


# yapf: disable
def enumerate(collection):
    """ Create a list of lists in which each element of the list is:
    [index, element]. For example, if you parse a FeatureCollection with 3
    Features you'll get: [[0, feat0], [1, feat1], [2, feat2]]
    :param collection: can be an ImageCollection or a FeatureCollection
    :return: ee.Collection
    """
    collist = collection.toList(collection.size())

    # first element
    ini = ee.Number(0)
    first_image = ee.Image(collist.get(0))
    first = ee.List([ini, first_image])

    start = ee.List([first])
    rest = collist.slice(1)

    def over_list(im, s):
        im = ee.Image(im)
        s = ee.List(s)
        last = ee.List(s.get(-1))
        last_index = ee.Number(last.get(0))
        index = last_index.add(1)
        return s.add(ee.List([index, im]))

    list = ee.List(rest.iterate(over_list, start))

    return list


def enumerateProperty(collection, name='enumeration'):
    """
    :param collection:
    :param name:
    :return:
    """
    enumerated = enumerate(collection)

    def over_list(ll):
        ll = ee.List(ll)
        index = ee.Number(ll.get(0))
        element = ll.get(1)
        return ee.Image(element).set(name, index)

    imlist = enumerated.map(over_list)
    return ee.ImageCollection(imlist)


def empty(value=0, names=None, from_dict=None):
    """ Create a constant image with the given band names and value, and/or
    from a dictionary of {name: value}
    :param names: list of names
    :type names: ee.List or list
    :param value: value for every band of the resulting image
    :type value: int or float
    :param from_dict: {name: value}
    :type from_dict: dict
    :rtype: ee.Image
    """
    image = ee.Image.constant(0)
    bandnames = ee.List([])
    if names:
        bandnames = names if isinstance(names, ee.List) else ee.List(names)

        def bn(name, img):
            img = ee.Image(img)
            newi = ee.Image(value).select([0], [name])
            return img.addBands(newi)

        image = ee.Image(bandnames.iterate(bn, image)) \
            .select(bandnames)

    if from_dict:
        from_dict = ee.Dictionary(from_dict)
        image = ee.Image(from_dict.toImage())

    if not from_dict and not names:
        image = ee.Image.constant(value)

    return image


def euclideanDistance(
    image1, image2, bands=None, discard_zeros=False,
    name='distance'
):
    """ Compute the Euclidean distance between two images. The image's bands
    is the dimension of the arrays.
    :param image1:
    :type image1: ee.Image
    :param image2:
    :type image2: ee.Image
    :param bands: the bands that want to be computed
    :type bands: list
    :param discard_zeros: pixel values equal to zero will not count in the
        distance computation
    :type discard_zeros: bool
    :param name: the name of the resulting band
    :type name: str
    :return: a distance image
    :rtype: ee.Image
    """
    if not bands:
        bands = image1.bandNames()

    image1 = image1.select(bands)
    image2 = image2.select(bands)

    proxy = empty(0, bands)
    image1 = proxy.where(image1.gt(0), image1)
    image2 = proxy.where(image2.gt(0), image2)

    if discard_zeros:
        # zeros
        zeros1 = image1.eq(0)
        zeros2 = image2.eq(0)

        # fill zeros with values from the other image
        image1 = image1.where(zeros1, image2)
        image2 = image2.where(zeros2, image1)

    a = image1.subtract(image2)
    b = a.pow(2)
    c = b.reduce('sum')
    d = c.sqrt()

    return d.rename(name)


def sumDistance(
    image, collection, bands=None, discard_zeros=False,
    name='sumdist'
):
    """ Compute de sum of all distances between the given image and the
    collection passed

    :param image:
    :param collection:
    :return:
    """
    condition = isinstance(collection, ee.ImageCollection)

    if condition:
        collection = collection.toList(collection.size())

    accum = ee.Image(0).rename(name)

    def over_rest(im, ini):
        ini = ee.Image(ini)
        im = ee.Image(im)
        dist = ee.Image(euclideanDistance(image, im, bands, discard_zeros)) \
            .rename(name)
        return ini.add(dist)

    return ee.Image(collection.iterate(over_rest, accum))


def removeIndex(list, index):
    """ Remove an element by its index """
    list = ee.List(list)
    index = ee.Number(index)
    size = list.size()

    def allowed():
        def zerof(list):
            return list.slice(1, list.size())

        def rest(list, index):
            list = ee.List(list)
            index = ee.Number(index)
            last = index.eq(list.size())

            def lastf(list):
                return list.slice(0, list.size().subtract(1))

            def restf(list, index):
                list = ee.List(list)
                index = ee.Number(index)
                first = list.slice(0, index)
                return first.cat(list.slice(index.add(1), list.size()))

            return ee.List(ee.Algorithms.If(last, lastf(list), restf(list, index)))

        return ee.List(ee.Algorithms.If(index, rest(list, index), zerof(list)))

    condition = index.gte(size).Or(index.lt(0))

    return ee.List(ee.Algorithms.If(condition, -1, allowed()))


def intersection(eelist, intersect):
    """ Find matching values. If ee_list1 has duplicated values that are
    present on ee_list2, all values from ee_list1 will apear in the result
    :param intersect: the other Earth Engine List
    :return: list with the intersection (matching values)
    :rtype: ee.List
    """
    eelist = ee.List(eelist)
    intersect = ee.List(intersect)
    newlist = ee.List([])

    def wrap(element, first):
        first = ee.List(first)

        return ee.Algorithms.If(
            intersect.contains(element),
            first.add(element), first
        )

    return ee.List(eelist.iterate(wrap, newlist))


def removeBands(image, bands):
    """ Remove the specified bands from an image """
    bnames = image.bandNames()
    bands = ee.List(bands)
    inter = intersection(bnames, bands)
    diff = bnames.removeAll(inter)
    return image.select(diff)


def replace(image, to_replace, to_add):
    """ Replace one band of the image with a provided band
    :param to_replace: name of the band to replace. If the image hasn't got
        that band, it will be added to the image.
    :type to_replace: str
    :param to_add: Image (one band) containing the band to add. If an Image
        with more than one band is provided, it uses the first band.
    :type to_add: ee.Image
    :return: Same Image provided with the band replaced
    :rtype: ee.Image
    """

    band = to_add.select([0])
    bands = image.bandNames()
    resto = bands.remove(to_replace)
    img_resto = image.select(resto)
    img_final = img_resto.addBands(band)
    return img_final


def medoidScore(
    collection, bands=None, discard_zeros=False,
    bandname='sumdist', normalize=False
):
    """ Compute a score to reflect 'how far' is from the medoid. Same params
     as medoid() """
    first_image = ee.Image(collection.first())
    if not bands:
        bands = first_image.bandNames()

    # Create a unique id property called 'enumeration'
    enumerated = enumerateProperty(collection)
    collist = enumerated.toList(enumerated.size())

    def over_list(im):
        im = ee.Image(im)
        n = ee.Number(im.get('enumeration'))

        # Remove the current image from the collection
        filtered = removeIndex(collist, n)

        # Select bands for medoid
        to_process = im.select(bands)

        def over_collist(img):
            return ee.Image(img).select(bands)

        filtered = filtered.map(over_collist)

        # Compute the sum of the euclidean distance between the current image
        # and every image in the rest of the collection
        dist = sumDistance(
            to_process, filtered,
            name=bandname,
            discard_zeros=discard_zeros
        )

        # Mask zero values
        if not normalize:
            # multiply by -1 to get the lowest value in the qualityMosaic
            dist = dist.multiply(-1)

        return im.addBands(dist)

    imlist = ee.List(collist.map(over_list))

    medcol = ee.ImageCollection.fromImages(imlist)

    # Normalize result to be between 0 and 1
    if normalize:
        min_sumdist = ee.Image(medcol.select(bandname).min()) \
            .rename('min_sumdist')
        max_sumdist = ee.Image(medcol.select(bandname).max()) \
            .rename('max_sumdist')

        def to_normalize(img):
            sumdist = img.select(bandname)
            newband = ee.Image().expression(
                '1-((val-min)/(max-min))',
                {
                    'val': sumdist,
                    'min': min_sumdist,
                    'max': max_sumdist
                }
            ).rename(bandname)
            return replace(img, bandname, newband)

        medcol = medcol.map(to_normalize)

    return medcol


def medoid(collection, bands=None, discard_zeros=True):
    """ Medoid Composite. Adapted from https://www.mdpi.com/2072-4292/5/12/6481
    :param collection: the collection to composite
    :type collection: ee.ImageCollection
    :param bands: the bands to use for computation. The composite will include
        all bands
    :type bands: list
    :param discard_zeros: Masked and pixels with value zero will not be use
        for computation. Improves dark zones.
    :type discard_zeros: bool
    :return: the Medoid Composite
    :rtype: ee.Image
    """
    medcol = medoidScore(collection, bands, discard_zeros)
    comp = medcol.qualityMosaic('sumdist')
    final = removeBands(comp, ['sumdist'])
    return final
# yapf: enable



"""
   Copyright 2021 Dugal Harris - dugalh@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from enum import Enum


class CompositeMethod(str, Enum):
    """
    Enumeration for the compositing method, i.e. the method for finding a composite pixel from the stack of
    corresponding input image pixels.
    """
    q_mosaic = 'q-mosaic'
    """ 
    Use the unmasked pixel with the highest cloud distance (distance to nearest cloud). Where more than one pixel has 
    the same cloud distance, the first one in the stack is selected.     
    """

    mosaic = 'mosaic'
    """ Use the first unmasked pixel in the stack. """

    medoid = 'medoid'
    """
    Use the medoid of the unmasked pixels.  This is the pixel from the image having the minimum summed difference (
    across bands) from the median of the image stack. Maintains the original relationship between bands. See 
    https://www.mdpi.com/2072-4292/5/12/6481 for detail.    
    """

    median = 'median'
    """ Use the median of the unmasked pixels. """

    mode = 'mode'
    """ Use the mode of the unmasked pixels. """

    mean = 'mean'
    """ Use the mean of the unmasked pixels. """


class CloudMaskMethod(str, Enum):
    """ Enumeration for the Sentinel-2 cloud masking method. """
    cloud_prob = 'cloud-prob'
    """ Threshold the corresponding image from the Sentinel-2 cloud probability collection. """

    qa = 'qa'
    """ Use the `QA60` quality assessment band. """


class ResamplingMethod(str, Enum):
    """ Enumeration for the resampling method. """
    near = 'near'
    """ Nearest neighbour. """

    bilinear = 'bilinear'
    """ Bilinear. """

    bicubic = 'bicubic'
    """ Bicubic. """

    average = 'average'
    """ Average (recommended for downsampling). """


class ExportType(str, Enum):
    """ Enumeration for the export type. """
    drive = 'drive'
    """ Export to Google Drive. """

    asset = 'asset'
    """ Export to an Earth Engine asset. """

    cloud = 'cloud'
    """ Export to Google Cloud Storage. """

